;; html-edit-display.el -- HTML display mode

;; This library is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; This library is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with library; see the file COPYING.  If not, write to the
;; Free Software Foundation, Inc., 59 Temple Place - Suite 330,
;; Boston, MA 02111-1307, USA.

(defvar html-edit-display-mode1 "all")
(defvar html-edit-display-mode2 "cdata")
(defconst html-edit-display-mode-list
  '(("all") ("script") ("cdata") ("element")))

(unless (symbol-plist 'html-edit-invisible)
  (setplist 'html-edit-invisible
	    '(invisible t
	      rear-nonsticky t
	      intangible t
	      read-only t)))
(unless (symbol-plist 'html-edit-intangible)
  (setplist 'html-edit-intangible
	    '(rear-nonsticky t
	      intangible t
	      read-only t)))

(require 'html-edit-utils)
(require 'html-edit-sexp)

(defun html-edit-display-mode (mode)
  (interactive (list (html-edit-completing-read
		      "Display mode: "
		      html-edit-display-mode-list)))
  (unless (assoc mode html-edit-display-mode-list)
    (error "Unknown display mode %s" mode))
  (cond ((string= mode "all")    (html-edit-display-all))
	((string= mode "script") (html-edit-display-script))
	((string= mode "cdata")  (html-edit-display-cdata))
	((string= mode "element")(html-edit-display-toggle-element))
	))

(defun html-edit-display-toggle ()
  (interactive)
  (cond ((string= "all" html-edit-display-mode1)
	 (html-edit-display-mode html-edit-display-mode2))
	((string= "element" html-edit-display-mode1)
	 (html-edit-display-mode "element"))
	(t (setq html-edit-display-mode2 html-edit-display-mode1)
	   (html-edit-display-mode "all"))))

(defun html-edit-display-hide-region (beg end &optional remark)
  (when (< beg end)
    (put-text-property beg end
		       'category 'html-edit-invisible)
    (when remark
      (overlay-put (make-overlay beg end) 'before-string "..."))))

(defun html-edit-display-show-region (beg end)
  (remove-text-properties beg end
			  '(category html-edit-invisible))
  (remove-text-properties beg (min (1+ end) (point-max))
			  '(category html-edit-intangible))
  (html-edit-delete-overlay-region beg end))

(defun html-edit-display-all ()
  (interactive)
  (setq html-edit-display-mode1 "all")
  (let ((modified (buffer-modified-p))
	(inhibit-read-only t))
    (html-edit-display-show-region (point-min) (point-max))
    (restore-buffer-modified-p modified)))

(defun html-edit-display-script ()
  (interactive)
  (save-excursion
    (let ((modified (buffer-modified-p))
	  (inhibit-read-only t)
	  beg end last-pos)
      (html-edit-display-all)
      (setq html-edit-display-mode1 "script")
      (setq beg (point-min))
      (goto-char beg)
      (while (setq end (re-search-forward "<[?%]" nil t))
	(html-edit-display-hide-region beg end t)
	(setq beg (re-search-forward "[?%]>"))
	(setq beg (- beg 2))
	(setq last-pos beg))
      (if last-pos
	  (html-edit-display-hide-region last-pos (point-max) t)
	(error "PHP or ASP script is none"))
      (restore-buffer-modified-p modified)
      )))

(defun html-edit-display-cdata ()
  (interactive)
  (save-excursion
    (let ((modified (buffer-modified-p))
	  (inhibit-read-only t)
	  beg end)
      (html-edit-display-all)
      (setq html-edit-display-mode1 "cdata")

      (goto-char (point-min))
      (while (search-forward "<" nil t)
	(setq beg (1- (point))
	      end nil)
	(cond ((looking-at "/?[a-z]")
	       (setq end (html-edit-search-forward-tag-stop)))
	      ((= (char-after) ?%)
	       (search-forward "%>")
	       (setq end (point)))
	      ((= (char-after) ??)
	       (search-forward "?>")
	       (setq end (point)))
	      ((looking-at "!--")
	       (search-forward "-->")
	       (setq end (point)))
	      ((looking-at "!")
	       (search-forward ">")
	       (setq end (point))))
	(when end
	  (html-edit-display-hide-region beg end)
	  ))
      (restore-buffer-modified-p modified))))

(defun html-edit-display-check-element (start end)
  (text-property-any start end 'category 'html-edit-invisible))

(defun html-edit-display-beginning-of-element ()
  (if (looking-at "</[a-z]")
      (html-edit-backward-sexp)
    (when (not (looking-at "<[a-z]"))
      (html-edit-search-backward-tag)
      (if (looking-at "</[a-z]")
	  (html-edit-backward-sexp)))))

(defun html-edit-display-toggle-element ()
  (interactive)
  (setq html-edit-display-mode1 "element")
  (let ((modified (buffer-modified-p))
	pos
	elem-pos
	attr-pos
	content-pos
	intangible-beg
	intangible-end)
    (save-excursion
      (html-edit-safe-scan
       (setq pos (html-edit-element-positions))
       (setq elem-pos (nth 0 pos))
       (setq attr-pos (nth 1 pos))
       (setq content-pos (nth 2 pos))

       (if (html-edit-display-check-element (car elem-pos) (cdr elem-pos))
	   (html-edit-display-show-region (car elem-pos)
					  (min (1+ (cdr elem-pos))
					       (point-max)))

	 ;; intangible ΰ (ݥ󥿤ߤޤʤΰ) Υå
	 (when (or attr-pos content-pos)
	   (setq intangible-beg (if attr-pos
				    (car attr-pos)
				  (1- (car content-pos)))
		 intangible-end (if content-pos
				    (1+ (cdr content-pos))
				  (1+ (cdr attr-pos))))
	   (setq intangible-beg (max intangible-beg (point-min)))
	   (setq intangible-end (min intangible-end (point-max)))
	   (when (< intangible-beg intangible-end)
	     (put-text-property intangible-beg intangible-end
				'category 'html-edit-intangible)))

	 (when attr-pos
	   (html-edit-display-hide-region
	    (car attr-pos) (cdr attr-pos) (not content-pos)))
	 (when content-pos
	   (html-edit-display-hide-region
	    (car content-pos) (cdr content-pos) t))
	 )))
    (restore-buffer-modified-p modified)))

(provide 'html-edit-display)
