;; html-edit-mode.el --- Major mode for HTML
;; Copyright (C) 2002 Masanao Izumo
;; Maintainer: Masanao Izumo <iz@onicos.co.jp>
;; Keywords: HTML
;; Version: 1.5.0

;; This library is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; This library is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with library; see the file COPYING.  If not, write to the
;; Free Software Foundation, Inc., 59 Temple Place - Suite 330,
;; Boston, MA 02111-1307, USA.

;;; ~/.emacs Example:
;;; (autoload 'html-edit-mode "html-edit-mode" "Major mode for HTML" t)
;;; (setq auto-mode-alist
;;;      (append '(("\\.html?$"     . html-edit-mode)
;;;                ("\\.asp$"       . html-edit-mode)
;;;                ("\\.php$"       . html-edit-mode))))
;;;
;;;
;;; M-x html-edit-mode
;;; start HTML mode
;;;
;;; M-x html-edit-check
;;; check html syntax


(if (< (string-to-number emacs-version) 21)
    (error "Sorry, Emacs Version %s is not supported." emacs-version))

(defconst html-edit-version 1.500)
(defvar html-edit-rule-list
  '(("IE"	. ("html-edit-rule-ie"))
    ("NN"	. ("html-edit-rule-nn"))
    ("H4T"	. ("html-edit-rule-h4t"))
    ("IE+NN"	. ("html-edit-rule-ie"
		   "html-edit-rule-nn"))
    ("IMODE40"	. ("html-edit-rule-imode40"))
    ("IE+NN+H4T". ("html-edit-rule-ie"
		   "html-edit-rule-nn"
		   "html-edit-rule-h4t"))
    ))

(require 'html-edit-vars)
(require 'html-edit-check)
(require 'html-edit-rule)
(require 'html-edit-sexp)
(require 'html-edit-display)
(require 'html-edit-color)


(defvar html-edit-max-tag-length 256)

(defvar html-edit-current-rule "")
(defun html-edit-set-rule (rule)
  (interactive
   (list (html-edit-completing-read
	  "HTML Ruleset: "
	  html-edit-rule-list
	  t
	  )))
  (let (files)
    (setq files (cdr (assoc rule html-edit-rule-list)))
    (html-edit-initialize-element)
    (message "Loading rule %s..." rule)
    (setq html-edit-current-rule rule)
    (mapc #'(lambda (file)
	      (load file)
	      (html-edit-clear-entify))
	  files)
    (html-edit-set-attlist-property "META" "http-equiv" 'read-handler
				    (symbol-function
				     'html-edit-meta-http-equiv-read))
    (html-edit-set-attlist-property "META" "content" 'read-handler
				    (symbol-function
				     'html-edit-meta-content-read))
    (mapc #'(lambda (target)
	      (html-edit-set-attlist-property
	       (car target) (cdr target) 'read-handler
	       (symbol-function 'html-edit-url-read)))
	  html-edit-read-file-name-target)

    (mapc #'(lambda (target)
	      (html-edit-set-attlist-property
	       (car target) (cdr target) 'read-handler
	       (symbol-function 'html-edit-mime-type-read-handler)))
	  html-edit-mime-type-target)

    (mapc #'(lambda (target)
	      (html-edit-set-attlist-property
	       (car target) (cdr target) 'read-handler
	       (symbol-function 'html-edit-color-read)))
	  html-edit-color-target)

    (run-hooks 'html-edit-mode-rule-hook)
    (message "Loading rule %s...done" rule)
    ))

(defvar html-edit-mode-syntax-table nil
  "HTML ⡼ɤǤΥ󥿥åơ֥")
(if html-edit-mode-syntax-table
    ()
  (setq html-edit-mode-syntax-table (copy-syntax-table text-mode-syntax-table))
  (modify-syntax-entry ?.  "w" html-edit-mode-syntax-table)
  (modify-syntax-entry ?-  "w" html-edit-mode-syntax-table)
  (modify-syntax-entry ?<  "(>" html-edit-mode-syntax-table)
  (modify-syntax-entry ?>  ")<" html-edit-mode-syntax-table)
  (modify-syntax-entry ?\"  "\"    " html-edit-mode-syntax-table)
  (modify-syntax-entry ?\\  "\\   " html-edit-mode-syntax-table)
  )

(defvar html-edit-mode-map nil)
(if html-edit-mode-map
    ()
  (setq html-edit-mode-map (make-sparse-keymap))
  (define-key html-edit-mode-map "\M-\C-f" 'html-edit-forward-sexp)
  (define-key html-edit-mode-map "\M-\C-b" 'html-edit-backward-sexp)
  (define-key html-edit-mode-map "\C-x\C-s" 'html-edit-save-buffer)
  (define-key html-edit-mode-map "\C-x\C-c" 'html-edit-save-buffers-kill-emacs)
  (define-key html-edit-mode-map "\C-l" 'html-edit-recenter)
  (define-key html-edit-mode-map "\C-a" 'html-edit-beginning-of-line)
  (define-key html-edit-mode-map "\C-e" 'html-edit-end-of-line)
  (define-key html-edit-mode-map "\C-c?" 'html-edit-check)
  (define-key html-edit-mode-map "\C-he" 'html-edit-describe-element)
  (define-key html-edit-mode-map "\C-cd" 'html-edit-display-mode)
  (define-key html-edit-mode-map "\C-c\C-d\C-a" 'html-edit-display-all)
  (define-key html-edit-mode-map "\C-c\C-d\C-s" 'html-edit-display-script)
  (define-key html-edit-mode-map "\C-c\C-d\C-c" 'html-edit-display-cdata)
  (define-key html-edit-mode-map "\C-c\C-d\C-e"
    'html-edit-display-toggle-element)
  (define-key html-edit-mode-map "\C-c\t" 'html-edit-display-toggle)
  (define-key html-edit-mode-map "\C-c\C-m" 'html-edit-element-mark)
  (define-key html-edit-mode-map "\M-\t" 'html-edit-complete)

  ;;
  ;; Insert char entities
  ;;
  (define-key html-edit-mode-map "\C-c<"
    #'(lambda () (interactive) (insert "&lt;")))
  (define-key html-edit-mode-map "\C-c>"
    #'(lambda () (interactive) (insert "&gt;")))
  (define-key html-edit-mode-map "\C-c&"
    #'(lambda () (interactive) (insert "&amp;")))
  (define-key html-edit-mode-map "\C-c "
    #'(lambda () (interactive) (insert "&nbsp;")))
  (define-key html-edit-mode-map "\C-c\""
    #'(lambda () (interactive) (insert "&quot;")))
  )

(defconst html-edit-font-lock-keywords
  '(("<\\(/?[[:alpha:]][[:alnum:]]*\\)" 1 font-lock-function-name-face)
    ("&[[:alnum:]]+;?" . font-lock-variable-name-face)
    ("<!--\\(.\\|\n\\)*?-->" . font-lock-comment-face)
    ))


(defun html-edit-mode ()
  (interactive)
  (kill-all-local-variables)
  (use-local-map html-edit-mode-map)
  (setq mode-name "HTML Edit")
  (setq major-mode 'html-edit-mode)

  (set-syntax-table html-edit-mode-syntax-table)

  (make-local-variable 'font-lock-defaults)
  (setq font-lock-defaults
	'((html-edit-font-lock-keywords)
	  nil
	  t))

  (make-local-variable 'comment-start)
  (setq comment-start "<!--")
  (make-local-variable 'comment-end)
  (setq comment-end "-->")

;  (make-local-variable 'html-edit-element-obarray)
;  (make-local-variable 'html-edit-attlist-obarray)
;  (make-local-variable 'html-edit-entity-obarray)

  (unless (string= html-edit-current-rule html-edit-default-rule)
    (html-edit-set-rule html-edit-default-rule))

  (setq html-edit-mime-type-list
	(sort
	 (html-edit-list-to-set
	  (append html-edit-mime-type-list
		  (html-edit-read-mime-types-file
		   html-edit-mime-type-filename-list)))
	 'string<))

  (run-hooks 'html-edit-mode-hook)
  )

;;
;; Хåե򥻡֤˥å롣
;;
(defun html-edit-save-buffer (&optional args)
  (interactive "p")
  (let ((save-flag t))
    (if (and (buffer-modified-p) html-edit-check-syntax-savep)
	(condition-case err
	    (html-edit-check)
	  (error
	   (message "%s" err)
	   (ding)
	   (setq save-flag
		 (yes-or-no-p
		  (format "%s: Saving file? " err))))))
    (when save-flag
      (save-buffer args))))

(defun html-edit-save-buffers-kill-emacs (&optional arg)
  (interactive "P")
  (let ((save-flag t))
    (if (and (buffer-modified-p) html-edit-check-syntax-savep)
	(condition-case err
	    (html-edit-check)
	  (error
	   (message "%s" err)
	   (ding)
	   (setq save-flag
		 (yes-or-no-p
		  (format "%s: Saving file? " err)))
	   (if save-flag
	       (save-buffer))
	   )))
    (unless save-flag
      (set-buffer-modified-p nil)
      (kill-buffer (current-buffer)))
    (save-buffers-kill-emacs arg)))

(defun html-edit-recenter (&optional count)
  (interactive "P")
  (recenter count)
  (font-lock-fontify-buffer))

(defun html-edit-beginning-of-line ()
  (interactive)
  (html-edit-safe-scan
   (while (progn
	    (beginning-of-line)
	    (if (or (bobp)
		    (not (eq (get-text-property (1- (point)) 'category)
			     'html-edit-invisible)))
		nil
	      (backward-char)
	      t)))))

(defun html-edit-end-of-line ()
  (interactive)
  (html-edit-safe-scan
   (while (progn
	    (end-of-line)
	    (if (or (eobp)
		    (not (eq (get-text-property (point) 'category)
			     'html-edit-invisible)))
		nil
	      (forward-char)
	      t)))))

(defun html-edit-read-element-name-arround ()
  (save-excursion
    (cond ((looking-at "</?\\([a-z0-9]+\\)")
	   (match-string-no-properties 1))
	  ((looking-at "[a-z0-9]")
	   (skip-chars-backward "A-Za-z0-9")
	   (when (equal (char-before) ?/)
	       (backward-char))
	   (when (equal (char-before) ?<)
	       (backward-char))
	   (when (looking-at "</?\\([a-z0-9]+\\)")
	     (match-string-no-properties 1))))))

(defun html-edit-describe-element (element)
  (interactive (let ((completion-ignore-case t))
		 (list (html-edit-completing-read
			"HTML Element: "
			html-edit-element-obarray
			t
			(html-edit-read-element-name-arround)
			))))
  (with-output-to-temp-buffer "*Help*"
    (let ((attr (html-edit-attr-name-map (html-edit-get-attlist element)))
	  (elem (html-edit-get-element element))
	  (namelen 0)
	  buff)

      (princ "Rule set: ")
      (princ html-edit-current-rule)
      (princ "\n")

      (princ "Element name: ")
      (princ element)
      (princ "\n")

      (princ "Start tag: ")
      (if (nth 0 elem)
	  (princ "Required\n")
	(princ "Optional\n"))

      (princ "End tag:   ")
      (if (nth 1 elem)
	  (princ "Required\n")
	(if (eq (nth 2 elem) 'EMPTY)
	    (princ "None\n")
	  (princ "Optional\n")))

      (princ "Include elements: ")
      (princ (nth 2 elem))
      (princ "\n")


      (princ "Exclude elements: ")
      (princ (nth 3 elem))
      (princ "\n")

      (mapc #'(lambda (pair)
		(if (< namelen (length (car pair)))
		    (setq namelen (length (car pair)))))
	    attr)
      (setq namelen (+ namelen 8))
      (setq namelen (- namelen (% namelen 8)))

      (princ "\nAttributes:\n")
      (mapc #'(lambda (pair)
		(let* ((name (car pair))
		       (len (length name)))
		  (princ name)
		  (princ (make-string (- namelen len) ?\ ))
		  (princ (cdr pair))
		  (princ "\n")))
	    attr)
      (print-help-return-message))))

(defun html-edit-element-mark ()
  (interactive)
  (let (element-pos start end)
    (html-edit-safe-scan
     (setq element-pos (car (html-edit-element-positions)))
     (setq start (car element-pos))
     (setq end (cdr element-pos))
     (goto-char start)
     (set-mark-command nil)
     (goto-char end)
     (sit-for 1)
     (exchange-point-and-mark))))

;; (element-name attribute-name attribute-value char-entity beg end)
;;  0		 1		2		3	    4	5
(defun html-edit-parse-completion-target ()
  (let ((search-limit (max (point-min) (- (point) html-edit-max-tag-length)))
	elem-name
	elem-pos
	attr-name
	attr-value
	attr-value-pos
	ret
	beg
	end)
    (save-excursion
      ;; <element... Υå
      ;; &name... Υå
      (save-excursion
	(skip-chars-backward "A-Za-z0-9")

	(if (equal (char-before) ?&)
	    (progn
	      (looking-at "\\([[:alnum:]]*;?\\)")
	      (setq beg (match-beginning 0))
	      (setq end (match-end 0))
	      (setq ret (list nil nil nil
			      (concat "&" (buffer-substring-no-properties
					   beg end))
			      beg end)))

	  (and (equal (char-before) ?/)
	       (backward-char))
	  (when (and (equal (char-before) ?<)
		     (looking-at "/?\\([[:alnum:]]*\\)"))
	    (setq beg (match-beginning 1))
	    (setq end (match-end 1))
	    (setq ret (list (upcase (buffer-substring-no-properties beg end))
			    nil
			    nil
			    nil
			    beg
			    end)))))

      ;; <element... õ
      (setq elem-pos (html-edit-check-in-the-tag search-limit))
      (if elem-pos
	  (save-excursion
	    (goto-char elem-pos)
	    (if (looking-at "<\\([[:alpha:]][[:alnum:]]*\\)")
		(progn
		  (setq search-limit (match-end 1))
		  (setq elem-name (upcase (match-string-no-properties 1))))
	      (setq ret (make-list 6 nil)))) ; tag is not found
	(setq ret (make-list 6 nil))) ; tag is not found

      ;; name="value"  "value" ʬΥå
      (when (not ret)
	(save-excursion
	  (when (and (re-search-backward "[<>\"]" search-limit t)
		     (equal (char-after) ?\")
		     (equal (char-before) ?=))
	    (looking-at "\"\\([^<>\r\n\"]*\\)")
	    (setq beg (match-beginning 1))
	    (setq end (match-end 1))
	    (setq attr-value (buffer-substring-no-properties beg end))
	    (setq attr-value-pos (1- (point))))))

      ;; name=value  "value" ʬΥå
      (when (and (not ret) (not attr-value))
	(save-excursion
	  (skip-chars-backward "A-Za-z0-9./_-")
	  (when (equal (char-before) ?=)
	    (looking-at "[a-z0-9./_-]*")
	    (setq beg (match-beginning 0))
	    (setq end (match-end 0))
	    (setq attr-value (buffer-substring-no-properties beg end))
	    (setq attr-value-pos (1- (point))))))

      ;; name=...  "name" ʬΥå
      (when (not ret)
	(when attr-value-pos
	  (goto-char attr-value-pos))
	(skip-chars-backward "A-Za-z0-9-")
	(when (or (equal (char-before) ?\ )
		  (equal (char-before) ?\t)
		  (equal (char-before) ?\r)
		  (equal (char-before) ?\n))
	  (looking-at "[a-z0-9-]*")
	  (setq attr-name (downcase (match-string-no-properties 0)))
	  (unless attr-value-pos
	    (setq beg (match-beginning 0))
	    (setq end (match-end 0)))))

      (when (and (not ret) elem-name attr-name)
	(setq ret (list elem-name attr-name attr-value nil beg end)))
      )
    ret))

(defun html-edit-complete-collection-entity ()
  (mapcar #'(lambda (entry)
	      (list (car entry)
		    (concat (make-string (- 12 (length (car entry))) ?\ )
			    (nth 1 entry)
			    "\t"
			    (nth 2 entry))
		    ))
	  html-edit-character-entities))

(defun html-edit-complete-collection-attr-value (element attr-name attr-value)
  (let (value handler)
    (when element
      (setq value (html-edit-get-attlist element attr-name)))
    (when value
      (setq handler (html-edit-get-attlist-property element attr-name
						    'read-handler)))
    (if handler
	(funcall handler element attr-name value attr-value)
      (if (listp value)
	  (mapcar #'list value)))))

(defun html-edit-complete-collection-attr-name (element)
  (if element
      (html-edit-attr-name-map (html-edit-get-attlist element))
    nil))

(defun html-edit-complete-collection-element ()
  html-edit-element-obarray)

(defun html-edit-complete ()
  (interactive)
  (let ((completion-ignore-case t)
	(context (html-edit-parse-completion-target))
	pattern
	collection
	collection2
	target
	beg end)

    (when context
      (setq beg (nth 4 context))
      (setq end (nth 5 context))
      (goto-char end))

    (setq collection
	  (cond ((nth 3 context)
		 (setq pattern (nth 3 context))
		 (setq target 'char-entity)
		 (html-edit-complete-collection-entity))
		((nth 2 context)
		 (setq pattern (nth 2 context))
		 (setq target 'attr-value)
		 (html-edit-complete-collection-attr-value (nth 0 context)
							   (nth 1 context)
							   (nth 2 context)))
		((nth 1 context)
		 (setq pattern (nth 1 context))
		 (setq target 'attr-name)
		 (html-edit-complete-collection-attr-name (nth 0 context)))
		((nth 0 context)
		 (setq pattern (nth 0 context))
		 (setq target 'element)
		 (html-edit-complete-collection-element))
		(t nil)))

    (cond ((null collection)
	   (ding))
	  ((stringp collection)
	   ;; replace
	   (setq beg (nth 4 context)
		 end (nth 5 context))
	   (delete-region beg end)
	   (insert collection))
	  (t
	   (setq beg (nth 4 context)
		 end (nth 5 context))
	   (setq collection2 (try-completion pattern collection))

	   (cond ((eq collection2 t))
		 ((null collection2)
		  (message "Can't find completion for \"%s\"" pattern)
		  (ding))
		 ((not (string= pattern collection2))
		  (delete-region beg end)
		  (insert collection2))
		 (t
		  (let ((list (all-completions pattern collection)))
		    (when (eq target 'char-entity)
		      (setq list (mapcar #'(lambda (x)
					     (assoc x collection)) list)))
		    (with-output-to-temp-buffer "*Completions*"
		      (display-completion-list list))
		    (save-current-buffer
		      (set-buffer (get-buffer "*Completions*"))
		      (make-local-variable 'truncate-lines)
		      (setq truncate-lines t))
		    )))))))

(defun html-edit-meta-http-equiv-read (elem-name attr-name dtd-value
						 &optional initial-value)
  (let ((completion-ignore-case t))
    (html-edit-completing-read
     "META http-equiv: "
     (mapcar #'list html-edit-meta-http-equiv-list)
     nil
     initial-value)))

(defun html-edit-meta-content-read (elem-name attr-name dtd-value
					      &optional initial-value)
  (let ((completion-ignore-case t)
	(pos (point))
	(tag-begin (html-edit-check-in-the-tag
		    (max (point-min) (- (point) html-edit-max-tag-length))))
	mode
	prompt
	)
    (save-excursion
      (goto-char tag-begin)
      (when (search-forward "http-equiv=\"content-type\"" pos t)
	(setq mode 'content-type))

      (when (not mode)
	(goto-char tag-begin)
	(when (search-forward "http-equiv=\"content-script-type\"" pos t)
	  (setq prompt "Content-Script-Type: ")
	  (setq mode 'mime-type)))

      (when (not mode)
	(goto-char tag-begin)
	(when (search-forward "http-equiv=\"content-style-type\"" pos t)
	  (setq prompt "Content-Style-Type: ")
	  (setq mode 'mime-type))))
    (cond
     ((eq mode 'content-type)
      (let (mime-type charset)
	(if (and initial-value
		 (string-match "\\([a-z0-9.+-]+/[a-z0-9.+-]+\\)\\(;[ \r\n\t]*charset=\\([a-z0-9_-]+\\)\\)?"
			       initial-value))
	    (setq mime-type (match-string-no-properties 1 initial-value)
		  charset   (match-string-no-properties 3 initial-value))
	  (setq mime-type "text/html"))

	  (setq charset
		(completing-read "Charset: "
				 (mapcar #'list html-edit-charset-list)
				 nil
				 nil
				 charset))

	  (if (equal charset "")
	      mime-type
	    (concat mime-type "; charset=" charset))
	  ))
     ((eq mode 'mime-type)
      (html-edit-mime-type-read prompt initial-value)))))

(defun html-edit-url-read (elem-name attr-name dtd-value
				     &optional initial-value)
  (or initial-value
      (setq initial-value ""))
  (let ((insert-default-directory nil))
    (read-file-name
     (concat elem-name " " attr-name ": ")
     nil initial-value nil initial-value)))

(defun html-edit-read-mime-types-file (fname-or-list)
  (if (listp fname-or-list)
      (apply #'append (mapcar #'html-edit-read-mime-types-file fname-or-list))
    (let (ret)
      (unwind-protect
	  (save-excursion
	    (set-buffer (find-file-noselect fname-or-list t t))
	    (goto-char (point-min))
	    (while (re-search-forward "^\\([a-z0-9.+-]+/[a-z0-9.+-]+\\)" nil t)
	      (setq ret (cons (match-string-no-properties 1) ret)))
	    ))
      ret)))

(defun html-edit-mime-type-read (prompt &optional initial-value)
  (let ((completion-ignore-case t))
    (html-edit-completing-read
     prompt
     (mapcar #'list html-edit-mime-type-list)
     nil
     initial-value)))

(defun html-edit-mime-type-read-handler (elem-name attr-name dtd-value
						   &optional initial-value)
  (html-edit-mime-type-read (concat elem-name " " attr-name ": ")
			    initial-value))

(defun html-edit-color-read (elem-name attr-name dtd-value
				       &optional initial-value)
  (let ((completion-ignore-case t)
	rule colors)
    (setq rule (concat html-edit-current-rule "+"))
    (cond ((string-match "IE\\+" rule)
	   (setq colors html-edit-color-ie))
	  ((string-match "NN\\+" rule)
	   (setq colors html-edit-color-nn))
	  (t
	   (setq colors html-edit-color-basic)))
    (setq colors (append colors html-edit-color-216))
    (html-edit-completing-read
     (concat elem-name " " attr-name ": ")
     (mapcar #'list colors)
     nil
     initial-value)))

(provide 'html-edit-mode)
